using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Net;
using System.Threading.Tasks;

using Smdn.Net.Imap4;
using Smdn.Net.Imap4.Client;

namespace Smdn.Applications.SimpleMailer {
  public class ImapAccount : AccountBase {
    public ImapClient Client {
      get { return client; }
    }

    public bool ReadOnly {
      get; private set;
    }

    public override bool IsReady {
      get { return !(client == null || client.IsBusy); }
    }

    public ImapAccount(string name,
                       Uri serverUri,
                       ICredentialsByHost credential,
                       int timeout,
                       bool useTlsIfAvailable,
                       bool readOnly,
                       ISynchronizeInvoke synchronizingObject)
      : base(name, serverUri)
    {
      this.client = new ImapClient(serverUri);
      this.client.SynchronizingObject = synchronizingObject;
      this.credential = credential;

      this.client.Profile.Timeout = timeout;
      this.client.Profile.UseTlsIfAvailable = useTlsIfAvailable;

      this.ReadOnly = readOnly;
    }

    public override Task EnsureConnected()
    {
      if (client.IsConnected) {
        var tcs = new TaskCompletionSource<object>();

        tcs.SetResult(null);

        return tcs.Task;
      }

      return client.ConnectAsync(credential).ContinueWith((_) => {
        this.credential = null;
      }, TaskContinuationOptions.OnlyOnRanToCompletion);
    }

    private ImapClient client;
    private ICredentialsByHost credential;
  }
}
