using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Diagnostics;
using System.Drawing;
using System.Drawing.Drawing2D;
using System.IO;
using System.Net;
using System.Net.Security;
using System.Security.Cryptography.X509Certificates;
using System.Text;
using System.Threading;
using System.Threading.Tasks;
using System.Windows.Forms;

#if false
using Smdn.Collections;
#else
using System.Linq;
#endif
using Smdn.IO;
using Smdn.Formats;
using Smdn.Formats.Ini;
using Smdn.Formats.Mime;
using Smdn.Formats.Mime.Formatting;
using Smdn.Net.Imap4;
using Smdn.Net.Imap4.Client;
using Smdn.Net;
using Smdn.Net.Pop3;
using Smdn.Net.Pop3.Client;
using Smdn.Windows.Forms;

namespace Smdn.Applications.SimpleMailer {
  public partial class MainWindow {
    const string configFile = "SimpleMailer.ini";
    const string acceptedCertHashFile = "AcceptedCerts.ini";

    public MainWindow()
    {
      ImapSslConnection.ServerCertificateValidationCallback += ValidateServerCertificate;
      PopSslConnection.ServerCertificateValidationCallback += ValidateServerCertificate;
#if false
      ImapWebRequestDefaults.ClientID["version"] = System.Reflection.Assembly.GetExecutingAssembly().GetName().Version.ToString();
#endif

      InitializeComponent();
      Initialize();
    }

    private static class Fonts {
      public static readonly Font BoldFont;
      public static readonly Font StrikeoutFont;
      public static readonly Font Monospace;

      static Fonts()
      {
        BoldFont = new Font(Control.DefaultFont, FontStyle.Bold);
        StrikeoutFont = new Font(Control.DefaultFont, FontStyle.Strikeout);
        Monospace = new Font(FontFamily.GenericMonospace, Control.DefaultFont.SizeInPoints, FontStyle.Regular, GraphicsUnit.Point);
      }
    }

    private void Initialize()
    {
      InitializeContextMenu();

      toolStripMenuItemFile_Quit.Click += delegate { Close(); };

      toolStripMenuItemMailbox_Update.Click       += delegate(object sender, EventArgs e) { UpdateNode((sender as TreeViewContextMenuItem).ContextNode); };
      toolStripMenuItemMailbox_CancelDelete.Click += delegate(object sender, EventArgs e) { CancelDelete((sender as TreeViewContextMenuItem).ContextNode as PopMailboxNode); };
      toolStripMenuItemMailbox_Append.Click       += delegate(object sender, EventArgs e) { AppendMessage((sender as TreeViewContextMenuItem).ContextNode as ImapMailboxNode); };
      toolStripMenuItemMailbox_Create.Click       += delegate(object sender, EventArgs e) { CreateMailbox((sender as TreeViewContextMenuItem).ContextNode); };
      toolStripMenuItemMailbox_Delete.Click       += delegate(object sender, EventArgs e) { DeleteMailbox((sender as TreeViewContextMenuItem).ContextNode as ImapMailboxNode); };
      toolStripMenuItemMailbox_MoveToTrash.Click  += delegate(object sender, EventArgs e) { MoveMailboxToTrash((sender as TreeViewContextMenuItem).ContextNode as ImapMailboxNode); };
      toolStripMenuItemMailbox_Rename.Click       += delegate(object sender, EventArgs e) { RenameMailbox((sender as TreeViewContextMenuItem).ContextNode as ImapMailboxNode); };
      toolStripMenuItemMailbox_Subscribe.Click    += delegate(object sender, EventArgs e) { ShowSubscribeDialog((sender as TreeViewContextMenuItem).ContextNode); };
      toolStripMenuItemMailbox_Unsubscribe.Click  += delegate(object sender, EventArgs e) { UnsubscribeMailbox((sender as TreeViewContextMenuItem).ContextNode as ImapMailboxNode); };

      toolStripMenuItemMessage_Fetch.Click              += delegate(object sender, EventArgs e) { FetchAndSaveMessage((sender as ListViewContextMenuItem).ContextMessage); };
      toolStripMenuItemMessage_ToggleSeen.Click         += delegate(object sender, EventArgs e) { ToggleMessageSeen((sender as ListViewContextMenuItem).ContextMessage as ImapMessageItem); };
      toolStripMenuItemMessage_Expunge.Click            += delegate(object sender, EventArgs e) { ExpungeMessage((sender as ListViewContextMenuItem).ContextMessage); };
      toolStripMenuItemMessage_MoveToTrash.Click        += delegate(object sender, EventArgs e) { MoveMessageToTrash((sender as ListViewContextMenuItem).ContextMessage as ImapMessageItem); };
      toolStripMenuItemMessage_ToggleKeywords_Append.Click
                                                        += delegate(object sender, EventArgs e) { StoreMessageKeyword((sender as ListViewContextMenuItem).ContextMessage as ImapMessageItem); };

      toolStripMenuItemMessage_SaveAttachedFile.DropDownOpening += SaveAttachedFileDropDownOpening;
      toolStripMenuItemMessage_CopyTo.DropDownOpening += CopyOrMoveMessageToDropDownOpening;
      toolStripMenuItemMessage_MoveTo.DropDownOpening += CopyOrMoveMessageToDropDownOpening;

      toolStripMenuItemMessage_ToggleKeywords.DropDownOpening += ToggleMessageKeywordsDropDownOpening;

      dropDownListSections.SelectedIndexChanged += delegate(object sender, EventArgs e) {
        SectionSelected((sender as SectionDropDownList).SelectedItem);
      };
      buttonFetchAndSaveSection.Click += delegate {
        FetchAndSaveSection(dropDownListSections.SelectedItem);
      };

      treeViewMailbox.BeforeSelect += AccountOrMailboxBeforeSelect;
      treeViewMailbox.AfterSelect += AccountOrMailboxAfterSelect;

      listViewMessages.View = View.Details;
      listViewMessages.FullRowSelect = true;
      listViewMessages.HeaderStyle = ColumnHeaderStyle.Clickable;
      listViewMessages.Columns.Add("タイトル", 160);
      listViewMessages.Columns.Add("UID", 40);
      listViewMessages.Columns.Add("サイズ", 60);
      listViewMessages.Columns.Add("Content-Type", 80);
      listViewMessages.Columns.Add("受信日 (internal date)", 120);
      listViewMessages.Columns.Add("送信日", 120);
      listViewMessages.Columns.Add("送信者", 100);
      listViewMessages.Columns.Add("フラグ", 200);
      listViewMessages.SelectedIndexChanged += MessageSelected;
      listViewMessages.ColumnClick += delegate(object sender, ColumnClickEventArgs e) {
        var listView = sender as ListView;
        var sorter = new MessageListViewItemComparer(listView.ListViewItemSorter as MessageListViewItemComparer, e.Column);

        listView.ListViewItemSorter = sorter;
      };

      textBoxMessage.Multiline = true;
      textBoxMessage.ScrollBars = ScrollBars.Both;
      textBoxMessage.Font = Fonts.Monospace;

      textBoxStatus.ReadOnly = true;
      textBoxStatus.Multiline = true;
      textBoxStatus.ScrollBars = ScrollBars.Both;
      textBoxStatus.Font = Fonts.Monospace;

      textBoxServerInfo.ReadOnly = true;
      textBoxServerInfo.Multiline = true;
      textBoxServerInfo.ScrollBars = ScrollBars.Both;
      textBoxServerInfo.Font = Fonts.Monospace;

      timerUpdate.Interval = 5 * 60 * 1000; // 5 min
      timerUpdate.Tick += delegate {
        NoOp();
      };
      timerUpdate.Enabled = true;

      imapDownloadProgress = new Progress<ImapMessageFetchProgressInfo>(p => {
        ReportDownloadProgress(p.FetchedOctetCount, p.TotalOctetCount);
      });

      popDownloadProgress = new Progress<PopMessageRetrieveProgressInfo>(p => {
        ReportDownloadProgress(p.RetrievedOctetCount, p.TotalOctetCount);
      });
    }

    private ToolStripItem[] mailboxToolStripItems;
    private ToolStripItem[] messageToolStripItems;

    private void SetMailboxContextMenuProperty(Point? cursorPosition)
    {
      var contextNode = cursorPosition.HasValue
        ? treeViewMailbox.GetNodeAt(treeViewMailbox.PointToClient(cursorPosition.Value))
        : treeViewMailbox.SelectedNode;
      var account = (contextNode as IAccountContainer).Account;
      var isImap = (contextNode is ImapMailboxNode) || (contextNode is ImapAccountNode) || (contextNode is ImapNonExistentMailboxNode);
      var mailboxSelected = contextNode is ImapMailboxNode;
      var mailbox = mailboxSelected ? (contextNode as ImapMailboxNode).Mailbox : null;

      toolStripMenuItemMailbox_Append.Enabled =
      toolStripMenuItemMailbox_MoveToTrash.Enabled =
      toolStripMenuItemMailbox_Delete.Enabled =
      toolStripMenuItemMailbox_Rename.Enabled =
      toolStripMenuItemMailbox_Unsubscribe.Enabled = mailboxSelected;

      toolStripMenuItemMailbox_Append.Visible =
      toolStripMenuItemMailbox_Create.Visible =
      toolStripMenuItemMailbox_MoveToTrash.Visible =
      toolStripMenuItemMailbox_Delete.Visible =
      toolStripMenuItemMailbox_Rename.Visible =
      toolStripMenuItemMailbox_Subscribe.Visible =
      toolStripMenuItemMailbox_Unsubscribe.Visible = isImap;

      toolStripMenuItemMailbox_CancelDelete.Visible = !isImap && contextNode is IMailboxContainer;

      toolStripMenuItemMailbox_Append.ContextNode =
      toolStripMenuItemMailbox_MoveToTrash.ContextNode =
      toolStripMenuItemMailbox_Delete.ContextNode =
      toolStripMenuItemMailbox_Rename.ContextNode =
      toolStripMenuItemMailbox_Unsubscribe.ContextNode = contextNode;

      var accountSelected = contextNode is ImapAccountNode;

      toolStripMenuItemMailbox_CancelDelete.Enabled =
      toolStripMenuItemMailbox_Update.Enabled = contextNode is IAccountContainer ||
                                                contextNode is IMailboxContainer;

      toolStripMenuItemMailbox_Create.Enabled =
      toolStripMenuItemMailbox_Subscribe.Enabled = mailboxSelected || accountSelected;

      toolStripMenuItemMailbox_Create.Enabled &= (mailbox == null || mailbox.CanHaveChild);

      toolStripMenuItemMailbox_Create.Enabled |= (contextNode is ImapNonExistentMailboxNode);

      toolStripMenuItemMailbox_CancelDelete.ContextNode =
      toolStripMenuItemMailbox_Update.ContextNode =
      toolStripMenuItemMailbox_Create.ContextNode =
      toolStripMenuItemMailbox_Subscribe.ContextNode = contextNode;

      if (!account.IsReady) {
        toolStripMenuItemMailbox_Update.Enabled =
        toolStripMenuItemMailbox_CancelDelete.Enabled =
        toolStripMenuItemMailbox_Append.Enabled =
        toolStripMenuItemMailbox_Create.Enabled =
        toolStripMenuItemMailbox_Delete.Enabled =
        toolStripMenuItemMailbox_MoveToTrash.Enabled =
        toolStripMenuItemMailbox_Rename.Enabled =
        toolStripMenuItemMailbox_Subscribe.Enabled =
        toolStripMenuItemMailbox_Unsubscribe.Enabled = false;
      }
    }

    private void SetMessageContextMenuProperty()
    {
      var contextMessage     = (0 < listViewMessages.SelectedItems.Count) ? listViewMessages.SelectedItems[0] as IMessageContainer : null;
      var isImap = contextMessage is ImapMessageItem;
      var contextImapMessage = contextMessage as ImapMessageItem;

      toolStripMenuItemMessage_SaveAttachedFile.Enabled =
      toolStripMenuItemMessage_Fetch.Enabled = 
      toolStripMenuItemMessage_Expunge.Enabled = (contextMessage != null && !contextMessage.IsDeleted);

      toolStripMenuItemMessage_MoveTo.Enabled = 
      toolStripMenuItemMessage_CopyTo.Enabled = 
      toolStripMenuItemMessage_ToggleSeen.Enabled = 
      toolStripMenuItemMessage_MoveToTrash.Enabled =
      toolStripMenuItemMessage_ToggleKeywords.Enabled = (contextImapMessage != null && !contextMessage.IsDeleted);

      toolStripMenuItemMessage_MoveTo.Visible = 
      toolStripMenuItemMessage_CopyTo.Visible = 
      toolStripMenuItemMessage_ToggleSeen.Visible = 
      toolStripMenuItemMessage_MoveToTrash.Visible = 
      toolStripMenuItemMessage_ToggleKeywords.Visible = isImap;

      toolStripMenuItemMessage_SaveAttachedFile.ContextMessage =
      toolStripMenuItemMessage_Fetch.ContextMessage = 
      toolStripMenuItemMessage_MoveTo.ContextMessage = 
      toolStripMenuItemMessage_CopyTo.ContextMessage = 
      toolStripMenuItemMessage_ToggleSeen.ContextMessage = 
      toolStripMenuItemMessage_Expunge.ContextMessage = 
      toolStripMenuItemMessage_MoveToTrash.ContextMessage = 
      toolStripMenuItemMessage_ToggleKeywords.ContextMessage = contextMessage;

      toolStripMenuItemMessage_SaveAttachedFile.DropDownItems.Clear();
      toolStripMenuItemMessage_SaveAttachedFile.DropDownItems.Add(contextMenuDummyItem);

      toolStripMenuItemMessage_MoveTo.DropDownItems.Clear();
      toolStripMenuItemMessage_MoveTo.DropDownItems.Add(contextMenuDummyItem);

      toolStripMenuItemMessage_CopyTo.DropDownItems.Clear();
      toolStripMenuItemMessage_CopyTo.DropDownItems.Add(contextMenuDummyItem);

      if (contextImapMessage == null)
        toolStripMenuItemMessage_ToggleSeen.Checked = false;
      else
        toolStripMenuItemMessage_ToggleSeen.Checked = contextImapMessage.Message.Flags.Contains(ImapMessageFlag.Seen);

      if (contextMessage != null && !contextMessage.Account.IsReady) {
        toolStripMenuItemMessage_SaveAttachedFile.Enabled =
        toolStripMenuItemMessage_Fetch.Enabled = 
        toolStripMenuItemMessage_MoveTo.Enabled = 
        toolStripMenuItemMessage_CopyTo.Enabled = 
        toolStripMenuItemMessage_ToggleSeen.Enabled = 
        toolStripMenuItemMessage_Expunge.Enabled = 
        toolStripMenuItemMessage_MoveToTrash.Enabled = 
        toolStripMenuItemMessage_ToggleKeywords.Enabled = false;
      }
    }

    private void InitializeContextMenu()
    {
      mailboxToolStripItems = new ToolStripItem[toolStripMenuItemMailbox.DropDownItems.Count];

      toolStripMenuItemMailbox.DropDownItems.CopyTo(mailboxToolStripItems, 0);

      messageToolStripItems = new ToolStripItem[toolStripMenuItemMessage.DropDownItems.Count];

      toolStripMenuItemMessage.DropDownItems.CopyTo(messageToolStripItems, 0);

      contextMenuMailbox.Items.Add(contextMenuDummyItem);

      contextMenuMailbox.Opening += delegate {
        SetMailboxContextMenuProperty(Cursor.Position);

        toolStripMenuItemMailbox.DropDownItems.Clear();
        contextMenuMailbox.Items.Clear();
        contextMenuMailbox.Items.AddRange(mailboxToolStripItems);
      };

      contextMenuMailbox.Closed += delegate {
        contextMenuMailbox.Items.Clear();
        contextMenuMailbox.Items.Add(contextMenuDummyItem);
        toolStripMenuItemMailbox.DropDownItems.Clear();
        toolStripMenuItemMailbox.DropDownItems.AddRange(mailboxToolStripItems);
      };

      toolStripMenuItemMailbox.DropDownOpening += delegate {
        SetMailboxContextMenuProperty(null);

        contextMenuMailbox.Items.Clear();
        contextMenuMailbox.Items.Add(contextMenuDummyItem);
        toolStripMenuItemMailbox.DropDownItems.Clear();
        toolStripMenuItemMailbox.DropDownItems.AddRange(mailboxToolStripItems);
      };

      toolStripMenuItemMailbox.DropDownClosed += delegate {
        toolStripMenuItemMailbox.DropDownItems.Clear();
        contextMenuMailbox.Items.Clear();
        contextMenuMailbox.Items.AddRange(mailboxToolStripItems);
      };


      contextMenuMessage.Items.Add(contextMenuDummyItem);

      contextMenuMessage.Opening += delegate {
        SetMessageContextMenuProperty();

        toolStripMenuItemMessage.DropDownItems.Clear();
        contextMenuMessage.Items.Clear();
        contextMenuMessage.Items.AddRange(messageToolStripItems);
      };

      contextMenuMessage.Closed += delegate {
        contextMenuMessage.Items.Clear();
        contextMenuMessage.Items.Add(contextMenuDummyItem);
        toolStripMenuItemMessage.DropDownItems.Clear();
        toolStripMenuItemMessage.DropDownItems.AddRange(messageToolStripItems);
      };

      toolStripMenuItemMessage.DropDownOpening += delegate {
        SetMessageContextMenuProperty();

        contextMenuMessage.Items.Clear();
        contextMenuMessage.Items.Add(contextMenuDummyItem);
        toolStripMenuItemMessage.DropDownItems.Clear();
        toolStripMenuItemMessage.DropDownItems.AddRange(messageToolStripItems);
      };

      toolStripMenuItemMessage.DropDownClosed += delegate {
        toolStripMenuItemMessage.DropDownItems.Clear();
        contextMenuMessage.Items.Clear();
        contextMenuMessage.Items.AddRange(messageToolStripItems);
      };
    }

    protected override void OnShown (EventArgs e)
    {
      base.OnShown (e);

      Application.DoEvents();

      LoadConfig();

      InitializeMailboxTree();
    }

    protected override void OnClosing (CancelEventArgs e)
    {
      base.OnClosing (e);

      if (!e.Cancel) {
        LogoutAll();
      }
    }

    private bool ValidateServerCertificate(object sender, X509Certificate certificate, X509Chain chain, SslPolicyErrors sslPolicyErrors)
    {
      if (InvokeRequired)
        return (bool)Invoke((Func<object, X509Certificate, X509Chain, SslPolicyErrors, bool>)ValidateServerCertificate, sender, certificate, chain, sslPolicyErrors);

      if (sslPolicyErrors == SslPolicyErrors.None)
        return true;
      else if (!verifyServerCertificates)
        return true;

      var acceptedCertHashes = File.Exists(acceptedCertHashFile) ? IniDocument.Load(acceptedCertHashFile) : new IniDocument();

      if (sslPolicyErrors == acceptedCertHashes.DefaultSection.Get(certificate.GetCertHashString(), EnumUtils.ParseIgnoreCase<SslPolicyErrors>))
        return true;

      using (var dialog = new ValidateCertDialog()) {
        switch (dialog.ValidateCert(this, certificate, chain, sslPolicyErrors)) {
          case ValidateCertDialog.Result.Accept:
            acceptedCertHashes.DefaultSection[certificate.GetCertHashString()] = sslPolicyErrors.ToString();
            acceptedCertHashes.Save(acceptedCertHashFile);
            return true;

          case ValidateCertDialog.Result.Abort:
            return false;

          case ValidateCertDialog.Result.AcceptOnlyOnce:
          default:
            return true;
        }
      }
    }

    private void LoadConfig()
    {
      Converter<string, bool> convertYesNo = (val) => val != "no";
      var config = IniDocument.Load(configFile);
      var accounts = new List<AccountBase>();

      var defaultSection = config.DefaultSection;

      mimeTypesFile = defaultSection.Get("MimeTypesFile", "/etc/mime.types");
      verifyServerCertificates = defaultSection.Get("VerifyServerCertificates", true, convertYesNo);

      // 証明書ファイルのセクション
      var certFilesSection = config["CertFiles"];

      foreach(var entry in certFilesSection) {
        try {
          ImapSslConnection.ClientCertificates.Add(new X509Certificate2(entry.Value));
          PopSslConnection.ClientCertificates.Add(new X509Certificate2(entry.Value));
        }
        catch (Exception ex) {
          AppendStatus("証明書{0}を読み込めませんでした", entry.Value);
          AppendStatus("Exception: {0}", ex);
        }
      }

      config.Remove(certFilesSection.Name);

      // フォントのセクション
      var fontsSection = config["Fonts"];

      try {
        textBoxMessage.Font = new Font(fontsSection.Get("Message_FontName"),
                                       fontsSection.Get("Message_FontSize", float.Parse),
                                       FontStyle.Regular,
                                       GraphicsUnit.Point);
      }
      catch {
        textBoxMessage.Font = Fonts.Monospace;
      }

      try {
        textBoxStatus.Font =
        textBoxServerInfo.Font = new Font(fontsSection.Get("Status_FontName"),
                                          fontsSection.Get("Status_FontSize", float.Parse),
                                          FontStyle.Regular,
                                          GraphicsUnit.Point);
      }
      catch {
        textBoxStatus.Font =
        textBoxServerInfo.Font = Fonts.Monospace;
      }

      config.Remove(fontsSection.Name);

      // アカウントのセクション
      foreach (var section in config) {
        if (section.IsDefaultSection)
          continue;

        try {
          var serverUri = section.GetThrowException<Uri>("ServerURL", ConvertUtils.ToUri);

          AccountBase account;

          if (ImapUri.IsImap(serverUri)) {
            var username = ImapStyleUriParser.GetUser(serverUri);
            var credential = section.Entries.ContainsKey("Password")
              ? (ICredentialsByHost)new NetworkCredential(username, section["Password"])
              : (ICredentialsByHost)new PasswordInputCredential(username, this);

            account = new ImapAccount(section.Name,
                                      serverUri,
                                      credential,
                                      section.Get("Timeout", System.Threading.Timeout.Infinite, Convert.ToInt32),
                                      section.Get("TryTLS", true, convertYesNo),
                                      section.Get("ReadOnly", true, convertYesNo),
                                      this);
          }
          else if (PopUri.IsPop(serverUri)) {
            var username = PopStyleUriParser.GetUser(serverUri);
            var credential = section.Entries.ContainsKey("Password")
              ? (ICredentialsByHost)new NetworkCredential(username, section["Password"])
              : (ICredentialsByHost)new PasswordInputCredential(username, this);

            account = new PopAccount(section.Name,
                                     serverUri,
                                     credential,
                                     section.Get("Timeout", System.Threading.Timeout.Infinite, Convert.ToInt32),
                                     section.Get("TryTLS", true, convertYesNo));
          }
          else {
            throw new UriFormatException("unsupported uri scheme");
          }

          accounts.Add(account);
        }
        catch (Exception ex) {
          AppendStatus("{0}のアカウント設定を読み込めませんでした", section.Name);
          AppendStatus("Exception: {0}", ex);
        }
      }

      this.imapAccounts = accounts.FindAll(delegate(AccountBase account) { return account is ImapAccount; })
                                  .ConvertAll(delegate(AccountBase account) { return account as ImapAccount; }).ToArray();
      this.popAccounts = accounts.FindAll(delegate(AccountBase account) { return account is PopAccount; })
                                 .ConvertAll(delegate(AccountBase account) { return account as PopAccount; }).ToArray();
      this.allAccounts = accounts.ToArray();

      if (this.allAccounts.Length == 0)
        AppendStatus("表示できるアカウント設定がありません。");
    }

    private void InitializeMailboxTree()
    {
      var tasks = new List<Task>();

      foreach (var account in imapAccounts) {
        tasks.Add(UpdateMailboxTree(account));
      }

      foreach (var account in popAccounts) {
        tasks.Add(UpdateMailboxTree(account));
      }

      //Task.WaitAll(tasks.ToArray());
    }

    private async Task UpdateMailboxTree(ImapAccount account)
    {
      await RunBusyOperation(async delegate {
        // このアカウント用のルートノードを作成する
        ImapAccountNode rootNode = null;

        try {
          foreach (var removing in imapMailboxNodeMap.Where(pair => pair.Value.Account == account).ToList()) {
            imapMailboxNodeMap.Remove(removing.Key);
          }

          treeViewMailbox.BeginUpdate();

          foreach (TreeNode node in treeViewMailbox.Nodes) {
            if ((node is ImapAccountNode) && (node as ImapAccountNode).Account == account) {
              rootNode = node as ImapAccountNode;
              break;
            }
          }

          if (rootNode == null) {
            rootNode = new ImapAccountNode(account);
            treeViewMailbox.Nodes.Add(rootNode);
          }
          else {
            // 子ノードをすべて削除
            rootNode.Nodes.Clear();
          }

          rootNode.ExpandAll();
        }
        finally {
          treeViewMailbox.EndUpdate();
        }

        // アカウントを接続状態にする
        try {
          rootNode.SetStatusText("接続中...");

          await account.EnsureConnected();
        }
        catch (Exception ex) {
          rootNode.SetStatusText("(接続できませんでした)");

          AppendStatus(ex, "{0}の接続に失敗しました", account.Name);

          return;
        }

        account.Client.ExistMessageCountChanged   += MailboxSizeChanged;
        account.Client.RecentMessageCountChanged  += MailboxSizeChanged;

        account.Client.MessageDeleted       += MessageStatusChanged;
        account.Client.MessageStatusChanged += MessageStatusChanged;

        // メールボックスのノードを作成する
        IEnumerable<ImapMailboxInfo> mailboxes = null;

        try {
          rootNode.SetStatusText("メールボックスを取得中...");

          const ImapMailboxListOptions options =
            ImapMailboxListOptions.SubscribedOnly | ImapMailboxListOptions.RequestStatus;

          mailboxes = await account.Client.GetMailboxesAsync(options);
        }
        catch (Exception ex) {
          rootNode.SetStatusText("(メールボックスを取得できませんでした)");

          AppendStatus(ex, "{0}のメールボックス一覧を取得できませんでした", account.Name);

          return;
        }

        try {
          treeViewMailbox.BeginUpdate();

          foreach (var mailbox in mailboxes.OrderBy(mailbox => mailbox.FullName, StringComparer.Ordinal)) {
            var node = new ImapMailboxNode(account, mailbox);

            if (string.Equals(mailbox.FullName, mailbox.Name, StringComparison.Ordinal)) {
              // top level mailbox
              rootNode.Nodes.Add(node);
            }
            else {
              // nested mailbox
              TreeNode parentNode = rootNode;
              var fullName = new StringBuilder();

              foreach (var hierarchy in mailbox.SplitMailboxHierarchy(mailbox.ParentMailboxName)) {
                TreeNode hierarchyNode = parentNode.Nodes.Cast<TreeNode>().Where(n => string.Equals(n.Name, hierarchy, StringComparison.Ordinal)).FirstOrDefault();

                fullName.Append(hierarchy);

                if (hierarchyNode == null) {
                  hierarchyNode = new ImapNonExistentMailboxNode(account, hierarchy, fullName.ToString());
                  parentNode.Nodes.Add(hierarchyNode);
                }

                parentNode = hierarchyNode;

                fullName.Append(mailbox.MailboxSeparator);
              }

              parentNode.Nodes.Add(node);
            }

            imapMailboxNodeMap.Add(mailbox, node);
          }
        }
        finally {
          rootNode.SetStatusText(null);

          rootNode.ExpandAll();

          treeViewMailbox.EndUpdate();
        }
      });

      treeViewMailbox.Refresh();
    }

    private async Task UpdateMailboxTree(PopAccount account)
    {
      await RunBusyOperation(async delegate {
        // このアカウント用のルートノードを作成する
        PopAccountNode rootNode = null;

        try {
          popMailboxNodeMap.Remove(account);

          treeViewMailbox.BeginUpdate();

          foreach (TreeNode node in treeViewMailbox.Nodes) {
            if ((node is PopAccountNode) && (node as PopAccountNode).Account == account) {
              rootNode = node as PopAccountNode;
              break;
            }
          }

          if (rootNode == null) {
            rootNode = new PopAccountNode(account);
            treeViewMailbox.Nodes.Add(rootNode);
          }
          else {
            // 子ノードをすべて削除
            rootNode.Nodes.Clear();
          }

          rootNode.ExpandAll();
        }
        finally {
          treeViewMailbox.EndUpdate();
        }

        // アカウントを接続状態にする
        try {
          rootNode.SetStatusText("接続中...");

          await account.EnsureConnected();
        }
        catch (Exception ex) {
          rootNode.SetStatusText("(接続できませんでした)");

          AppendStatus(ex, "{0}の接続に失敗しました", account.Name);

          return;
        }

        // メールボックスのノードを作成する
        var mailboxNode = new PopMailboxNode(account);

        rootNode.Nodes.Add(mailboxNode);

        popMailboxNodeMap[account] = mailboxNode;

        rootNode.SetStatusText(null);

        rootNode.ExpandAll();
      });

      treeViewMailbox.Refresh();
    }

    private async void UpdateNode(TreeNode node)
    {
      if (node is ImapMailboxNode) {
        await UpdateMailbox(node as ImapMailboxNode);
      }
      else if (node is ImapAccountNode) {
        await UpdateMailboxTree((node as ImapAccountNode).Account);
      }
      else if (node is PopMailboxNode) {
        await UpdateMailboxTree((node as PopMailboxNode).Account);
      }
      else if (node is PopAccountNode) {
        await UpdateMailboxTree((node as PopAccountNode).Account);
      }
    }

    private async Task UpdateMailbox(ImapMailboxNode node)
    {
      try {
        if (node.Mailbox.IsOpen) {
          await node.Account.Client.OpenedMailbox.RefreshAsync();
        }
        else {
          await node.Mailbox.RefreshAsync();

          node.UpdateText();
        }
      }
      catch (Exception ex) {
        AppendStatus(ex, "{0}を最新の状態に更新できませんでした", node.Mailbox.Name);
      }
    }

    private void MailboxSizeChanged(object sender, ImapMailboxSizeChangedEventArgs e)
    {
      foreach (var pair in imapMailboxNodeMap) {
        if (pair.Key.Client != sender)
          continue;

        if (pair.Key.FullName == e.Mailbox.FullName) {
          pair.Value.UpdateText();
          break;
        }
      }
    }

    private void MessageStatusChanged(object sender, ImapMessageStatusChangedEventArgs e)
    {
      var changedMessages = new List<ImapMessageInfo>(e.Messages);

      foreach (var item in listViewMessages.Items) {
        var i = item as ImapMessageItem;

        if (i == null)
          continue;

        if (changedMessages.Exists(delegate(ImapMessageInfo m) { return m == i.Message; })) {
          i.UpdateView();

          changedMessages.Remove(i.Message);

          if (changedMessages.Count == 0)
            return;
        }
      }
    }

    private void AccountOrMailboxBeforeSelect(object sender, TreeViewCancelEventArgs e)
    {
      var mailboxContainer = e.Node as IMailboxContainer;

      if (mailboxContainer != null && !mailboxContainer.Account.IsReady)
        e.Cancel = true;
    }

    private async void AccountOrMailboxAfterSelect(object sender, TreeViewEventArgs e)
    {
      var accountContainer = e.Node as IAccountContainer;

      if (e.Node is IMailboxContainer) {
        var mailboxContainer = e.Node as IMailboxContainer;

        ViewServerInfo(mailboxContainer);

        await UpdateMessageList(mailboxContainer);

        await UpdateQuotaUsage(mailboxContainer);

        AppendStatus(mailboxContainer.GetStatusText());
      }
      else if (accountContainer != null) {
        ViewServerInfo(accountContainer);
      }
    }

    private async Task UpdateMessageList(IMailboxContainer container)
    {
      if (container.Account is ImapAccount)
        await UpdateMessageList(container.Account as ImapAccount, container.GetMailbox() as ImapMailboxInfo);
      else
        await UpdateMessageList(container.Account as PopAccount);

      dropDownListSections.Items.Clear();
      dropDownListSections.Enabled = false;
    }

    private async Task UpdateMessageList(ImapAccount account, ImapMailboxInfo mailbox)
    {
      if (!mailbox.Exists || mailbox.IsUnselectable) {
        AppendStatus("{0}は選択できません", mailbox.Name);
        return;
      }
      else if (mailbox == account.Client.OpenedMailbox) {
        return;
      }

      var mailboxNode = imapMailboxNodeMap[mailbox];

      // メッセージの一覧を取得
      AppendStatus("{0}のメッセージの一覧を取得しています", mailbox.Name);

      try {
        try {
          listViewMessages.Enabled = false;

          if (account.Client.OpenedMailbox != null)
            await account.Client.OpenedMailbox.CloseAsync();

          await account.Client.OpenMailboxAsync(mailbox, account.ReadOnly);

          mailboxNode.UpdateText();

          toolStripProgressBarProgress.Minimum = 0;
          toolStripProgressBarProgress.Maximum = (int)account.Client.OpenedMailbox.ExistMessageCount;
          toolStripProgressBarProgress.Value = 0;
          toolStripProgressBarProgress.Step = 1;
          toolStripProgressBarProgress.Style = ProgressBarStyle.Blocks;

          CreateMessageList(account, await account.Client.OpenedMailbox.GetMessages(ImapMessageFetchAttributeOptions.AllAttributes).ToListAsync());

          toolStripProgressBarProgress.Value = 0;
        }
        finally {
          listViewMessages.Enabled = true;
        }

        AppendStatus("全{0}件のメッセージがあります",account.Client.OpenedMailbox.ExistMessageCount);
      }
      catch (Exception ex) {
        AppendStatus(ex, "{0}のメッセージ一覧を取得できませんでした", mailbox.Name);
      }
    }

    private void CreateMessageList(ImapAccount account, IEnumerable<ImapMessageInfo> messages)
    {
      try {
        listViewMessages.BeginUpdate();
        listViewMessages.Items.Clear();

        foreach (var message in messages) {
          listViewMessages.Items.Add(new ImapMessageItem(account, message, imapDownloadProgress));

          toolStripProgressBarProgress.Increment(1);
        }
      }
      finally {
        listViewMessages.EndUpdate();
      }
    }

    private async Task UpdateMessageList(PopAccount account)
    {
      // メッセージの一覧を取得
      AppendStatus("{0}のメッセージの一覧を取得しています", account.Name);

      try {
        toolStripProgressBarProgress.Minimum = 0;
        toolStripProgressBarProgress.Maximum = (int)account.Client.MessageCount;
        toolStripProgressBarProgress.Value = 0;
        toolStripProgressBarProgress.Step = 1;
        toolStripProgressBarProgress.Style = ProgressBarStyle.Blocks;

        var messages = await account.Client.GetMessagesAsync(true).ContinueWith(t => {
          return t.Result.Select(delegate(PopMessageInfo info) {
            toolStripProgressBarProgress.Increment(1);

            return Tuple.Create(info, info.ReadAs<Mail>(0, Mail.Load));
          });
        });

        CreateMessageList(account, messages);

        toolStripProgressBarProgress.Value = 0;

        AppendStatus("全{0}件のメッセージがあります", account.Client.MessageCount);
      }
      catch (Exception ex) {
        AppendStatus(ex, "{0}のメッセージ一覧を取得できませんでした", account.Name);
      }
    }

    private void CreateMessageList(PopAccount account, IEnumerable<Tuple<PopMessageInfo, Mail>> messages)
    {
      try {
        listViewMessages.BeginUpdate();
        listViewMessages.Items.Clear();

        foreach (var message in messages) {
          listViewMessages.Items.Add(new PopMessageItem(account, message.Item1, message.Item2, popDownloadProgress));
        }
      }
      finally {
        listViewMessages.EndUpdate();
      }
    }

    private async Task UpdateQuotaUsage(IMailboxContainer container)
    {
      var mailbox = container.GetMailbox() as ImapMailboxInfo;

      if (mailbox == null)
        return;

      var removingItems = new List<QuotaResourceToolStripItem>();

      foreach (var item in statusStripMain.Items) {
        var quotaResourceItem = item as QuotaResourceToolStripItem;

        if (quotaResourceItem != null)
          removingItems.Add(quotaResourceItem);
      }

      foreach (var item in removingItems) {
        statusStripMain.Items.Remove(item);
      }

      if (!mailbox.Exists)
        return;

      AppendStatus("{0}のクォータ使用率を取得しています", mailbox.Name);

      try {
        foreach (var quota in await mailbox.GetQuotaAsync()) {
          foreach (var resource in quota.Resources) {
            var item = new QuotaResourceToolStripItem();

            item.Resource= resource;
            item.QuotaRoot = quota.Root;

            statusStripMain.Items.Add(item);
          }
        }

        AppendStatus("{0}のクォータ使用率を取得しました", mailbox.Name);
      }
      catch (Exception ex) {
        AppendStatus(ex, "{0}のクォータ使用率を取得できませんでした", mailbox.Name);
      }
    }

    private void MessageSelected(object sender, EventArgs e)
    {
      var container = (listViewMessages.SelectedItems.Count <= 0)
        ? null
        : listViewMessages.SelectedItems[0] as IMessageContainer;

      if (container is ImapMessageItem)
        UpdateSectionsDropDownList(container.Account, (container as ImapMessageItem).Message);
      else
        UpdateSectionsDropDownList(null, null);

      if (container == null)
        return;

      MessageSelected(container);
    }

    private CancellationTokenSource loadMailCancellationTokenSource = null;
    private ManualResetEventSlim loadMailWaitHandle = new ManualResetEventSlim(true);
    private IMessageContainer currentSelectedMessageContainer = null;

    private async void MessageSelected(IMessageContainer container)
    {
      if (currentSelectedMessageContainer != null) {
        if (currentSelectedMessageContainer == container)
          return; // already selected

        if (loadMailCancellationTokenSource != null && !currentSelectedMessageContainer.CanCancelLoading) {
          AppendStatus("{0}の本文を取得中です", currentSelectedMessageContainer.UniqueId);
          return;
        }
      }

      currentSelectedMessageContainer = container;

      if (currentSelectedMessageContainer.IsDeleted) {
        AppendStatus("{0}は削除されています", currentSelectedMessageContainer.UniqueId);
        return;
      }

      if (loadMailCancellationTokenSource != null) {
        loadMailCancellationTokenSource.Cancel();
        loadMailWaitHandle.Wait();
        loadMailWaitHandle.Reset();
      }

      loadMailCancellationTokenSource = new CancellationTokenSource();

      textBoxMessage.Text = "取得中...";
      textBoxMessage.Enabled = false;

      AppendStatus("{0}の本文を取得しています", currentSelectedMessageContainer.UniqueId);

      try {
        try {
          var sw = Stopwatch.StartNew();

          await currentSelectedMessageContainer.LoadMailAsync(loadMailCancellationTokenSource.Token);

          sw.Stop();

          ViewMessage(currentSelectedMessageContainer.Mail);

          AppendStatus("{0}の本文を取得しました ({1:N2}secs)", currentSelectedMessageContainer.UniqueId, sw.ElapsedMilliseconds / 1000.0);

          toolStripProgressBarProgress.Value = 0;

          textBoxMessage.Enabled = true;
        }
        catch (OperationCanceledException) {
          AppendStatus("{0}の本文を取得を中断しました", currentSelectedMessageContainer.UniqueId);

          textBoxMessage.Text = string.Empty;
        }
        catch (Exception ex) {
          AppendStatus(ex, "{0}の本文を取得または解析に失敗しました", currentSelectedMessageContainer.UniqueId);

          textBoxMessage.Text = string.Empty;
        }
      }
      finally {
        loadMailCancellationTokenSource = null;

        loadMailWaitHandle.Set();
      }
    }

    private void UpdateSectionsDropDownList(AccountBase account, ImapMessageInfo message)
    {
      var messagePart = (message == null) ? null : message.MessagePart;

      dropDownListSections.Enabled = (messagePart != null);

      dropDownListSections.BeginUpdate();

      try {
        dropDownListSections.Items.Clear();

        if (messagePart == null) {
          dropDownListSections.SelectedIndex = -1;
        }
        else {
          messagePart.Traverse(true, subpart => {
            dropDownListSections.Items.Add(new SectionDropDownListItem(account, subpart, imapDownloadProgress));
          });

          try {
            sectionSelectionLock = true;

            dropDownListSections.SelectedIndex = 0; // this will cause of calling SectionSelected
          }
          finally {
            sectionSelectionLock = false;
          }
        }
      }
      finally {
        dropDownListSections.EndUpdate();
      }
    }

    private bool sectionSelectionLock = false;

    private void SectionSelected(SectionDropDownListItem sectionItem)
    {
      if (sectionItem.IsDeleted) {
        AppendStatus("{0}は削除されています", sectionItem.UniqueId);
        return;
      }

      if (sectionSelectionLock)
        return;

      if (sectionItem.MessagePart.ContentType.TypeEqualsIgnoreCase("multipart") ||
          sectionItem.MessagePart.ContentType.TypeEqualsIgnoreCase("message")) {
        MessageSelected(sectionItem);
      }
      else if (sectionItem.MessagePart.ContentType.TypeEqualsIgnoreCase("text")) {
        string message = null;
        var sw = new Stopwatch();

        AppendStatus("{0}の本文を取得しています", sectionItem.UniqueId);

        try {
          sw.Start();

          try {
            message = sectionItem.MessagePart.ReadAllText(ImapMessageFetchBodyOptions.DecodeContent);
          }
          catch (Exception ex) {
            AppendStatus("本文のデコードに失敗しました");
            AppendStatus("Exception: {0}", ex);
          }

          sw.Stop();
        }
        catch (Exception ex) {
          AppendStatus(ex, "{0}の本文を取得できませんでした", sectionItem.UniqueId);
          return;
        }

        if (message != null)
          ViewMessage(message);

        AppendStatus("完了 ({0:N2}secs)", sw.ElapsedMilliseconds / 1000.0);
      }
      else {
        var sb = new StringBuilder(4096);

        sb.AppendLine("(非テキストメッセージです)");
        sb.AppendLine();

        sb.AppendFormat("Content-Type: {0}", sectionItem.MessagePart.ContentType).AppendLine();
        sb.AppendFormat("Length: {0}", sectionItem.MessagePart.Length).AppendLine();
        sb.AppendFormat("Encoding: {0}", sectionItem.MessagePart.ContentTransferEncoding).AppendLine();
        sb.AppendFormat("Disposition: {0} (filename: {1})",
                        sectionItem.MessagePart.ContentDisposition,
                        sectionItem.MessagePart.GetContentFileName()).AppendLine();

        ViewMessage(sb.ToString());
      }
    }

    private void ViewMessage(string message)
    {
      textBoxMessage.Text = message;
    }

    private void ViewMessage(Mail mail)
    {
      var sb = new StringBuilder(4096);

      if (mail.MimeType.TypeEqualsIgnoreCase("text") || mail.MimeType.TypeEqualsIgnoreCase("multipart"))
        sb.AppendLine(mail.ReadContentAsText());
      else
        sb.AppendLine("(非テキストメッセージです)");

      if (0 < mail.AttachedFiles.Count) {
        sb.AppendLine("[添付ファイル]");
        foreach (var attachedFile in mail.AttachedFiles) {
          sb.AppendFormat("{0} : {1} ({2})",
                          attachedFile.BaseMessage.PartSpecifier, 
                          attachedFile.FileName ?? "(ファイル名なし)",
                          attachedFile.MimeType);
          sb.AppendLine();
        }
      }

      textBoxMessage.Text = sb.ToString();
    }

    private void ViewServerInfo(IAccountContainer container)
    {
      if (container.Account is ImapAccount)
        ViewServerInfo(container.Account as ImapAccount);
      else if (container.Account is PopAccount)
        ViewServerInfo(container.Account as PopAccount);
    }

    private void ViewServerInfo(ImapAccount account)
    {
      if (!account.Client.IsConnected) {
        textBoxServerInfo.Text = "(接続されていません)";
        return;
      }

      var sb = new StringBuilder();

      sb.AppendLine("[server capabilities]");
      sb.AppendLine(string.Join(", ", account.Client.ServerCapabilities.ToStringArray()));
      sb.AppendLine();

      sb.AppendLine("[server ID]");

      foreach (var pair in account.Client.ServerID) {
        sb.AppendFormat("{0}: {1}", pair.Key, pair.Value);
        sb.AppendLine();
      }
      sb.AppendLine();

      var namespaces = account.Client.ServerNamespace;

      foreach (var ns in new[] {
        new {Namespaces = namespaces.PersonalNamespaces, Title = "personal"},
        new {Namespaces = namespaces.SharedNamespaces, Title = "shared"},
        new {Namespaces = namespaces.OtherUsersNamespaces, Title = "other users"},
      }) {
        sb.AppendFormat("[namespaces ({0})]", ns.Title);
        sb.AppendLine();
        foreach (var desc in ns.Namespaces) {
          sb.AppendFormat("  prefix: \"{0}\" delimiter: \"{1}\"", desc.Prefix, desc.HierarchyDelimiter);
          sb.AppendLine();
        }
        sb.AppendLine();
      }

      textBoxServerInfo.Text = sb.ToString();
    }

    private void ViewServerInfo(PopAccount account)
    {
      if (!account.Client.IsConnected) {
        textBoxServerInfo.Text = "(接続されていません)";
        return;
      }

      var sb = new StringBuilder();

      sb.AppendLine("[server capabilities]");
      sb.AppendLine(string.Join(", ", account.Client.ServerCapabilities.ToStringArray()));
      sb.AppendLine();

      textBoxServerInfo.Text = sb.ToString();
    }

    private readonly string fileDialogMessageFilterString = FileDialogFilter.CreateFilterString(new[] {
      new[] {"emlファイル", "*.eml"},
      new[] {"すべてのファイル", "*.*"},
    });

    private void AppendMessage(ImapMailboxNode node)
    {
      var mailbox = node.Mailbox;

      using (var openFileDialog = new OpenFileDialog()) {
        openFileDialog.Filter = fileDialogMessageFilterString;
        openFileDialog.Title = string.Format("{0}へファイルをアップロード", mailbox.Name);
        openFileDialog.Multiselect = false;

        if (openFileDialog.ShowDialog(this) != DialogResult.OK)
          return;

        if (PathUtils.AreExtensionEqual(openFileDialog.FileName, ".eml")) {
          using (var stream = openFileDialog.OpenFile()) {
            mailbox.AppendMessage(stream.Length, DateTimeOffset.Now, delegate(Stream writeStream) {
              stream.CopyTo(writeStream, 0x4000);
            });
          }
        }
        else {
          mailbox.AppendMessage(DateTimeOffset.Now, delegate(Stream writeStream) {
            WriteMimeCapsulatedMessageTo(openFileDialog.FileName, writeStream);
          });
        }

        /* TODO: append-uid
        if (appendedMessageUri != null) {
          var appendedMessage = account.Request(new Uri(appendedMessageUri, string.Format("?UID {0}", ImapStyleUriParser.GetUid(appendedMessageUri))), delegate(ImapWebResponse response) {
            return response.Messages[0];
          });

          listViewMessages.Items.Add(new ImapMessageItem(account, appendedMessage));
        }
        */

        MessageBox.Show(string.Format("{0}を{1}へアップロードしました", openFileDialog.FileName, mailbox.Name));
      }
    }

    private void WriteMimeCapsulatedMessageTo(string path, Stream stream)
    {
      var assemblyName = System.Reflection.Assembly.GetExecutingAssembly().GetName();
      var description = string.Format("{0}\r\nuploaded by {1}", path, assemblyName.Name);

      // TODO: Content-Type: *;format=flowed
      var part1 = new MimeMessageBuilder();

      part1.MimeType = MimeType.TextPlain;
      part1.WriteContent(description, ContentTransferEncodingMethod.SevenBit, Encoding.ASCII);

      MimeMessageBuilder part2;

      using (var attachmentStream = File.OpenRead(path)) {
        part2 = MimeMessageBuilder.CreateAttachment(attachmentStream,
                                                    MimeType.FindMimeTypeByExtension(path, mimeTypesFile) ?? MimeType.ApplicationOctetStream,
                                                    Path.GetFileName(path),
                                                    description,
                                                    true /*inline*/);
      }

      using (var message = MimeMessageBuilder.CreateMultipart(new[] {part1, part2})) {
        message.Headers.Add(new MimeHeaderBuilder("Date", DateTimeConvert.ToRFC822DateTimeString(DateTimeOffset.Now)));
        message.Headers.Add(new MimeHeaderBuilder("Subject", path, MimeEncodingMethod.Base64, Encoding.UTF8));
        message.Headers.Add(new MimeHeaderBuilder("User-Agent", string.Format("{0} {1}", assemblyName.Name, assemblyName.Version)));

        message.Save(stream);
      }
    }

    private async void CreateMailbox(TreeNode node)
    {
      ImapAccount account;
      ImapMailboxInfo parentMailbox;
      string createMailboxName = null;

      if (node is ImapMailboxNode) {
        account = (node as ImapMailboxNode).Account;
        parentMailbox = (node as ImapMailboxNode).Mailbox;
      }
      else if (node is ImapAccountNode) {
        account = (node as ImapAccountNode).Account;
        parentMailbox = null;
      }
      else if (node is ImapNonExistentMailboxNode) {
        account = (node as ImapNonExistentMailboxNode).Account;
        parentMailbox = null;
        createMailboxName = (node as ImapNonExistentMailboxNode).FullName;
      }
      else {
        return;
      }

      if (createMailboxName == null) {
        using (var dialog = new InputDialog()) {
          dialog.Text = "メールボックスの作成";
          dialog.Description = string.Format("新しく{0}に作成するメールボックスの名前を指定してください。",
                                             (parentMailbox == null) ? account.Name : parentMailbox.Name);
          dialog.InputText = "新しいメールボックス";

          if (dialog.ShowDialog(this) != DialogResult.OK)
            return;

          createMailboxName = dialog.InputText;
        }
      }

      try {
        if (parentMailbox == null)
          await account.Client.CreateMailboxAsync(createMailboxName, true);
        else
          await parentMailbox.CreateChildAsync(createMailboxName, true);

        AppendStatus("{0}を作成しました", createMailboxName);
      }
      catch (Exception ex) {
        AppendStatus(ex, "{0}の作成に失敗しました", createMailboxName);
      }

      await UpdateMailboxTree(account);
    }

    private async void DeleteMailbox(ImapMailboxNode node)
    {
      var mailbox = node.Mailbox;
      var result = MessageBox.Show(this,
                                   string.Format("{0}を削除します。　この操作は元に戻せません。", mailbox.Name),
                                   "メールボックスの削除", 
                                   MessageBoxButtons.OKCancel,
                                   MessageBoxIcon.Warning,
                                   MessageBoxDefaultButton.Button2);

      if (result != DialogResult.OK)
        return;

      try {
        await mailbox.DeleteAsync(true);

        AppendStatus("{0}を削除しました", mailbox.Name);
      }
      catch (Exception ex) {
        AppendStatus(ex, "{0}の削除に失敗しました", mailbox.Name);
      }

      await UpdateMailboxTree(node.Account);
    }

    private async void MoveMailboxToTrash(ImapMailboxNode node)
    {
      var account = node.Account;
      var trashMailbox = GetOrCreateTrashMailbox(account);

      if (trashMailbox == null)
        return;

      await node.Mailbox.MoveToAsync(trashMailbox, true);

      await UpdateMailboxTree(node.Account);
    }

    private ImapMailboxInfo GetOrCreateTrashMailbox(ImapAccount account)
    {
      foreach (var n in imapMailboxNodeMap.Values) {
        if (n.Account != account)
          continue;

        try {
          return account.Client.GetOrCreateMailbox("Trash", true);
        }
        catch (Exception ex) {
          AppendStatus(ex, "メールボックス'Trash'を作成できませんでした");
        }
      }

      return null; // failed
    }

    private async void RenameMailbox(ImapMailboxNode node)
    {
      string newMailboxName;

      using (var dialog = new InputDialog()) {
        dialog.Text = "メールボックスの名前の変更";
        dialog.Description = "新しいメールボックスの名前を入力してください。";
        dialog.InputText = node.Mailbox.FullName;

        if (dialog.ShowDialog(this) != DialogResult.OK)
          return;
        else if (dialog.InputText == node.Mailbox.FullName)
          return;

        newMailboxName = dialog.InputText;
      }

      try {
        await node.Mailbox.MoveToAsync(newMailboxName, true);

        AppendStatus("{0}の名前を{1}に変更しました", node.Mailbox.Name, newMailboxName);
      }
      catch (Exception ex) {
        AppendStatus(ex, "{0}の名前を{1}に変更できませんでした", node.Mailbox.Name, newMailboxName);
      }

      await UpdateMailboxTree(node.Account);
    }

    private async void ShowSubscribeDialog(TreeNode node)
    {
      ImapAccount account;

      if (node is ImapAccountNode)
        account = (node as ImapAccountNode).Account;
      else if (node is ImapMailboxNode)
        account = (node as ImapMailboxNode).Account;
      else
        return;

      using (var dialog = new SubscribeMailboxDialog(account)) {
        dialog.ShowDialog(this);
      }

      await UpdateMailboxTree(account);
    }

    private async void UnsubscribeMailbox(ImapMailboxNode node)
    {
      var mailbox = node.Mailbox;

      try {
        await mailbox.UnsubscribeAsync();

        AppendStatus("{0}を購読解除しました", mailbox.Name);
      }
      catch (Exception ex) {
        AppendStatus(ex, "{0}の購読解除に失敗しました", mailbox.Name);
      }

      await UpdateMailboxTree(node.Account);
    }

    private async void CancelDelete(PopMailboxNode node)
    {
      try {
        await node.Account.Client.CancelDeleteAsync();

        node.UpdateText();

        await UpdateMessageList(node.Account);

        AppendStatus("{0}の削除をキャンセルしました", node.Account.Name);
      }
      catch (Exception ex) {
        AppendStatus(ex, "{0}の削除キャンセルに失敗しました", node.Account.Name);
      }
    }

    private async void FetchAndSaveMessage(IMessageContainer message)
    {
      var subject = (message is ImapMessageItem) 
        ? (message as ImapMessageItem).Message.EnvelopeSubject
        : (message as PopMessageItem).Mail.Subject;

      using (var saveFileDialog = new SaveFileDialog()) {
        saveFileDialog.AddExtension = true;
        saveFileDialog.Filter = fileDialogMessageFilterString;
        saveFileDialog.Title = string.Format("{0}をファイルに保存", message.UniqueId);
        saveFileDialog.FileName = string.Format("{0}.eml", PathUtils.ReplaceInvalidFileNameChars(subject, " "));

        if (saveFileDialog.ShowDialog(this) != DialogResult.OK)
          return;

        AppendStatus("{0}を{1}に保存しています", message.UniqueId, saveFileDialog.FileName);

        try {
          // TODO: cancellation
          await message.SaveAsync(saveFileDialog.FileName, CancellationToken.None);

          AppendStatus("{0}を{1}に保存しました", message.UniqueId, saveFileDialog.FileName);

          MessageBox.Show(this, string.Format("{0}を{1}に保存しました", message.UniqueId, saveFileDialog.FileName));
        }
        catch (OperationCanceledException) {
          AppendStatus("{0}の保存を中断しました", message.UniqueId);
        }
        catch (Exception ex) {
          AppendStatus(ex, "{0}の保存に失敗しました", message.UniqueId);
        }
      }
    }

    private void FetchAndSaveSection(SectionDropDownListItem sectionItem)
    {
      if (sectionItem == null)
        return;

      var part = sectionItem.MessagePart;

      string extension;

      if (part.ContentType.TypeEqualsIgnoreCase("multipart") ||
          part.ContentType.TypeEqualsIgnoreCase("message")) {
        extension = ".eml";
      }
      else if (part.ContentType.TypeEqualsIgnoreCase("text")) {
        if (part.ContentType.SubTypeEqualsIgnoreCase("html"))
          extension = ".html";
        else
          extension = ".txt";
      }
      else {
        extension = string.Empty;
      }

      var filename = part.GetContentFileName();

      if (string.IsNullOrEmpty(filename))
        filename = "section" + part.Specifier + extension;

      filename = PathUtils.ReplaceInvalidFileNameChars(filename, " ");

      using (var saveFileDialog = new SaveFileDialog()) {
        saveFileDialog.Filter = fileDialogMessageFilterString;
        saveFileDialog.Title = string.Format("{0}をファイルに保存", sectionItem.UniqueId);
        saveFileDialog.FileName = filename;

        if (saveFileDialog.ShowDialog(this) != DialogResult.OK)
          return;

        sectionItem.MessagePart.Save(saveFileDialog.FileName, ImapMessageFetchBodyOptions.DecodeContent);

        MessageBox.Show(string.Format("{0}を{1}に保存しました", sectionItem.UniqueId, saveFileDialog.FileName));
      }
    }

    private void ToggleMessageSeen(ImapMessageItem item)
    {
      ToggleMessageKeyword(item, ImapMessageFlag.Seen);
    }

    private void ToggleMessageKeywordsDropDownOpening(object sender, EventArgs e)
    {
      var item = sender as ListViewContextMenuItem;

      item.DropDownItems.Clear();

      var mailbox = (treeViewMailbox.SelectedNode as ImapMailboxNode).Account.Client.OpenedMailbox;

      if (mailbox == null)
        return;

      foreach (ImapMessageFlag flag in mailbox.ApplicableFlags) {
        if (flag.IsSystemFlag)
          continue;

        var flagItem = new ListViewTargetFlagContextMenuItem();

        flagItem.Text = flag.Value;
        flagItem.ContextMessage = item.ContextMessage;
        flagItem.TargetFlag = flag;
        flagItem.Checked = (item.ContextMessage as ImapMessageItem).Message.Flags.Contains(flag);
        flagItem.Click += delegate(object clickSender, EventArgs clickEventArgs) {
          var i = clickSender as ListViewTargetFlagContextMenuItem;

          ToggleMessageKeyword(i.ContextMessage as ImapMessageItem, i.TargetFlag);
        };

        item.DropDownItems.Add(flagItem);
      }

      if (!mailbox.IsAllowedToCreateKeywords)
        return;

      toolStripMenuItemMessage_ToggleKeywords_Append.ContextMessage = item.ContextMessage;

      item.DropDownItems.Add(toolStripMenuItemMessage_ToggleKeywords_Append);
    }

    private void StoreMessageKeyword(ImapMessageItem item)
    {
      string keyword = null;

      using (var dialog = new InputDialog()) {
        dialog.Text = "フラグの追加";
        dialog.Description = "新しく設定するフラグの名前を指定してください。";

        /* TODO: validation
        dialog.Validating += delegate(object sender, CancelEventArgs e) {
          e.Cancel = !ImapMessageFlag.IsValidKeyword(dialog.InputText);
          MessageBox.Show("invalid keyword");
        };
        */

        if (dialog.ShowDialog(this) != DialogResult.OK)
          return;

        keyword = dialog.InputText;
      }

      try {
        var storeDataItem = ImapStoreDataItem.AddFlags(keyword);

        item.Message.Store(storeDataItem);

        AppendStatus("{0}に'{1}'を設定しました", item.UniqueId, storeDataItem);
      }
      catch (Exception ex) {
        AppendStatus(ex, "{0}に'{1}'を設定できませんでした", item.UniqueId, keyword);
      }
    }

    private void ToggleMessageKeyword(ImapMessageItem item, ImapMessageFlag keyword)
    {
      try {
        item.Message.ToggleFlags(keyword);

        AppendStatus("{0}に'{1}'を設定/設定解除しました", item.UniqueId, keyword);
      }
      catch (Exception ex) {
        AppendStatus(ex, "{0}に'{1}'を設定/設定解除できませんでした", item.UniqueId, keyword);
      }
    }

    private void ExpungeMessage(IMessageContainer message)
    {
      try {
        if (message is ImapMessageItem) {
          (message as ImapMessageItem).Message.Delete();
        }
        else {
          var popMessageItem = message as PopMessageItem;

          popMessageItem.MessageInfo.MarkAsDeleted();

          popMessageItem.UpdateView();

          popMailboxNodeMap[popMessageItem.Account].UpdateText();
        }

        AppendStatus("{0}を削除しました", message.UniqueId);
      }
      catch (Exception ex) {
        AppendStatus(ex, "{0}を削除できませんでした", message.UniqueId);
      }
    }

    private void CopyOrMoveMessageToDropDownOpening(object sender, EventArgs e)
    {
      var listViewContextMenuItem = sender as ListViewContextMenuItem;

      listViewContextMenuItem.DropDownItems.Clear();

      var messageItem = listViewContextMenuItem.ContextMessage as ImapMessageItem;

      if (messageItem == null)
        return;

      var copy = (sender == toolStripMenuItemMessage_CopyTo);

      foreach (var pair in imapMailboxNodeMap) {
        if (pair.Value.Account == messageItem.Account) {
          var dropDownItem = new ListViewTargetMailboxContextMenuItem();

          dropDownItem.Text = pair.Key.FullName;
          dropDownItem.ContextMessage = messageItem;
          dropDownItem.TargetMailbox = pair.Key;

          if (copy)
            dropDownItem.Click += delegate(object clickSender, EventArgs clickEvent) {
              var item = clickSender as ListViewTargetMailboxContextMenuItem;
              var contextMessageItem = item.ContextMessage as ImapMessageItem;

              CopyMessageTo(contextMessageItem, item.TargetMailbox);
            };
          else
            dropDownItem.Click += delegate(object clickSender, EventArgs clickEvent) {
              var item = clickSender as ListViewTargetMailboxContextMenuItem;
              var contextMessageItem = item.ContextMessage as ImapMessageItem;

              MoveMessageTo(contextMessageItem, item.TargetMailbox);
            };

          listViewContextMenuItem.DropDownItems.Add(dropDownItem);
        }
      }
    }

    private void MoveMessageToTrash(ImapMessageItem item)
    {
      var trashMailbox = GetOrCreateTrashMailbox(item.Account);

      if (trashMailbox == null)
        return;

      CopyOrMoveMessageTo(item, trashMailbox, false);
    }

    private void MoveMessageTo(ImapMessageItem item, ImapMailboxInfo mailbox)
    {
      CopyOrMoveMessageTo(item, mailbox, false);
    }

    private void CopyMessageTo(ImapMessageItem item, ImapMailboxInfo mailbox)
    {
      CopyOrMoveMessageTo(item, mailbox, true);
    }

    private void CopyOrMoveMessageTo(ImapMessageItem item, ImapMailboxInfo mailbox, bool copy)
    {
      try {
        if (copy) {
          item.Message.CopyTo(mailbox);
        }
        else {
          item.Message.MoveTo(mailbox);

          listViewMessages.Items.Remove(item);
        }

        AppendStatus("{0}を{1}にコピーしました", item.UniqueId, mailbox.Name);
      }
      catch (Exception ex) {
        AppendStatus(ex, "{0}を{1}にコピーできませんでした", item.UniqueId, mailbox.Name);
      }
    }

    private void SaveAttachedFileDropDownOpening(object sender, EventArgs e)
    {
      var listViewContextMenuItem = sender as ListViewContextMenuItem;

      listViewContextMenuItem.DropDownItems.Clear();

      var messageItem = listViewContextMenuItem.ContextMessage;

      if (messageItem == null)
        return;

      if (messageItem.Mail == null)
        return;

      if (messageItem.Mail.AttachedFiles.Count == 0) {
        var dropDownItem = new ListViewTargetAttachedFileContextMenuItem();

        dropDownItem.Text = "(添付ファイルはありません)";

        listViewContextMenuItem.DropDownItems.Add(dropDownItem);

        return;
      }

      foreach (var attachedFile in messageItem.Mail.AttachedFiles) {
        var dropDownItem = new ListViewTargetAttachedFileContextMenuItem();

        dropDownItem.AttachedFile = attachedFile;
        dropDownItem.Text = string.Format("{0} : {1} ({2})",
                                          attachedFile.BaseMessage.PartSpecifier,
                                          attachedFile.FileName ?? "(ファイル名なし)",
                                          attachedFile.MimeType);
        dropDownItem.ContextMessage = messageItem;

        dropDownItem.Click += delegate(object clickSender, EventArgs clickEvent) {
          var item = clickSender as ListViewTargetAttachedFileContextMenuItem;

          SaveAttachedFile(item.AttachedFile);
        };

        listViewContextMenuItem.DropDownItems.Add(dropDownItem);
      }
    }

    private void SaveAttachedFile(AttachedFile attachedFile)
    {
      using (var saveFileDialog = new SaveFileDialog()) {
        var title = attachedFile.FileName ?? "添付ファイル";
        var filters = new List<string[]>();
        var extensions = MimeType.FindExtensionsByMimeType(attachedFile.MimeType);

        if (extensions != null && 1 <= extensions.Length)
          filters.Add(ArrayExtensions.Concat(new[] {attachedFile.MimeType.ToString()}, extensions));

        filters.Add(new[] {"すべてのファイル", "*.*"});

        saveFileDialog.AddExtension = true;
        saveFileDialog.Filter = FileDialogFilter.CreateFilterString(filters.ToArray());
        saveFileDialog.Title = string.Format("{0}をファイルに保存", title);
        saveFileDialog.FileName = attachedFile.FileName;

        if (saveFileDialog.ShowDialog(this) != DialogResult.OK)
          return;

        attachedFile.SaveContent(saveFileDialog.FileName);

        MessageBox.Show(string.Format("{0}を{1}に保存しました", title, saveFileDialog.FileName));
      }
    }

    private void NoOp()
    {
      foreach (var account in imapAccounts) {
        account.Client.Refresh();
      }

      foreach (var account in popAccounts) {
        account.Client.KeepAlive();
      }
    }

    private void LogoutAll()
    {
      foreach (var account in imapAccounts) {
        try {
          account.Client.Logout();
        }
        catch (Exception ex) {
          AppendStatus("{0}からログアウト出来ませんでした", account.ServerUri);
          AppendStatus("Exception: {0}", ex);
        }
      }

      foreach (var account in popAccounts) {
        try {
          account.Client.Logout();
        }
        catch (Exception ex) {
          AppendStatus("{0}からログアウト出来ませんでした", account.ServerUri);
          AppendStatus("Exception: {0}", ex);
        }
      }
    }

    private async Task RunBusyOperation(Func<Task> action)
    {
      try {
        Cursor = Cursors.WaitCursor;

        await action();
      }
      finally {
        Cursor = Cursors.Default;
      }
    }

    private void AppendStatus(string format, params object[] args)
    {
      AppendStatus((Exception)null, format, args);
    }

    private void AppendStatus(Exception exception, string format, params object[] args)
    {
      var status = string.Format(format, args);

      toolStripStatusLabelStatusText.Text = status;

      statusStringBuffer.AppendLine(status);

      if (exception != null)
        statusStringBuffer.AppendFormat(exception.ToString());

      textBoxStatus.Text = statusStringBuffer.ToString();
      textBoxStatus.Select(textBoxStatus.TextLength - 1, 0);
      textBoxStatus.ScrollToCaret();
      textBoxStatus.Update();

      Console.Error.WriteLine(format, args);
    }

    private void ReportDownloadProgress(long current, long total)
    {
      var now = DateTime.Now;

      if (100 <= (now - progressLastUpdate).TotalMilliseconds)
        progressLastUpdate = now;
      else if (current < total)
        return;

      if (total == 0L) {
        toolStripProgressBarProgress.Maximum = 100;
        toolStripProgressBarProgress.Minimum = 0;
        toolStripProgressBarProgress.Value = 0;
        toolStripProgressBarProgress.Style = ProgressBarStyle.Continuous;
      }
      else {
        toolStripProgressBarProgress.Maximum = 100;
        toolStripProgressBarProgress.Minimum = 0;
        toolStripProgressBarProgress.Value = (int)((current * 100L) / total);
        toolStripProgressBarProgress.Style = ProgressBarStyle.Blocks;
      }
    }

    private class TreeViewContextMenuItem : ToolStripMenuItem {
      public TreeNode ContextNode {
        get; set;
      }
    }

    private class ListViewContextMenuItem : ToolStripMenuItem {
      public IMessageContainer ContextMessage {
        get; set;
      }
    }

    private class ListViewTargetMailboxContextMenuItem : ListViewContextMenuItem {
      public ImapMailboxInfo TargetMailbox {
        get; set;
      }
    }

    private class ListViewTargetFlagContextMenuItem : ListViewContextMenuItem {
      public ImapMessageFlag TargetFlag {
        get; set;
      }
    }

    private class ListViewTargetAttachedFileContextMenuItem : ListViewContextMenuItem {
      public AttachedFile AttachedFile {
        get; set;
      }
    }

    private class MessageListViewItemComparer : System.Collections.IComparer {
      public SortOrder SortOrder {
        get; private set;
      }

      public int Column {
        get; private set;
      }

      public MessageListViewItemComparer(MessageListViewItemComparer oldComparer, int column)
      {
        if (oldComparer != null && oldComparer.Column == column && oldComparer.SortOrder == SortOrder.Ascending)
          this.SortOrder = SortOrder.Descending;
        else
          this.SortOrder = SortOrder.Ascending;

        this.Column = column;
      }

      public int Compare(object x, object y)
      {
        var xx = (x as ListViewItem).SubItems[Column];
        var yy = (y as ListViewItem).SubItems[Column];

        int sign;

        switch (SortOrder) {
          case SortOrder.Ascending: sign = 1; break;
          case SortOrder.Descending: sign = -1; break;
          default: return 0;
        }

        var comp = (x is IMessageListViewItem)
          ? (x as IMessageListViewItem).CompareTo(Column, yy)
          : string.Compare(xx.Text, yy.Text);

        return sign * comp;
      }
    }

    private interface IAccountContainer {
      AccountBase Account { get; }
    }

    private class ImapAccountNode : TreeNode, IAccountContainer {
      public ImapAccount Account{
        get; private set;
      }

      AccountBase IAccountContainer.Account {
        get { return Account; }
      }

      public ImapAccountNode(ImapAccount account)
      {
        this.Account = account;

        Text = GetDefaultText();
      }

      public void SetStatusText(string statusText)
      {
        if (string.IsNullOrEmpty(statusText))
          Text = GetDefaultText();
        else
          Text = string.Concat(statusText, " ", GetDefaultText());
      }

      private string GetDefaultText()
      {
        return string.Format("{0} - {1}", Account.Name, Account.ServerUri.GetLeftPart(UriPartial.Authority));
      }
    }

    private class PopAccountNode : TreeNode, IAccountContainer {
      public PopAccount Account {
        get; private set;
      }

      AccountBase IAccountContainer.Account {
        get { return Account; }
      }

      public PopAccountNode(PopAccount account)
      {
        this.Account = account;

        Text = GetDefaultText();
      }

      public void SetStatusText(string statusText)
      {
        if (string.IsNullOrEmpty(statusText))
          Text = GetDefaultText();
        else
          Text = string.Concat(statusText, " ", GetDefaultText());
      }

      private string GetDefaultText()
      {
        return string.Format("{0} - {1}", Account.Name, Account.ServerUri.GetLeftPart(UriPartial.Authority));
      }
    }

    private interface IMailboxContainer : IAccountContainer {
      object GetMailbox();
      string GetStatusText();
    }

    private class ImapNonExistentMailboxNode : TreeNode, IAccountContainer {
      public ImapAccount Account {
        get; private set;
      }

      AccountBase IAccountContainer.Account {
        get { return Account; }
      }

      public string FullName {
        get; private set;
      }
     
      public ImapNonExistentMailboxNode(ImapAccount account, string name, string fullName)
      {
        this.Account = account;
        this.Text = name;
        this.Name = name;
        this.FullName = fullName;
        this.ForeColor = SystemColors.GrayText;
      }
    }

    private class ImapMailboxNode : TreeNode, IMailboxContainer {
      public ImapAccount Account {
        get; private set;
      }

      AccountBase IAccountContainer.Account {
        get { return Account; }
      }

      public ImapMailboxInfo Mailbox {
        get; set;
      }

      public ImapMailboxNode(ImapAccount account, ImapMailboxInfo mailbox)
      {
        this.Account = account;
        this.Mailbox = mailbox;

        this.Name = mailbox.Name;

        UpdateText();
      }

      public void UpdateText()
      {
        if (!Mailbox.Exists || Mailbox.IsUnselectable)
          ForeColor = SystemColors.GrayText;
        else if (0L < Mailbox.RecentMessageCount)
          ForeColor = Color.Blue;
        else
          ForeColor = Color.Black;

        Text = string.Format("{0} 新着{1} 未読{2} 全{3}",
                             Mailbox.Name,
                             Mailbox.RecentMessageCount,
                             Mailbox.UnseenMessageCount,
                             Mailbox.ExistMessageCount);
      }

      object IMailboxContainer.GetMailbox()
      {
        return Mailbox;
      }

      string IMailboxContainer.GetStatusText()
      {
        return string.Format("'{0}' 全{1}件 (新着{2}件 未読{3}件)",
                             Mailbox.FullName,
                             Mailbox.ExistMessageCount,
                             Mailbox.RecentMessageCount,
                             Mailbox.UnseenMessageCount);
      }
    }

    private class PopMailboxNode : TreeNode, IMailboxContainer {
      public PopAccount Account {
        get; private set;
      }

      AccountBase IAccountContainer.Account {
        get { return Account; }
      }

      public PopMailboxNode(PopAccount account)
      {
        this.Account = account;

        UpdateText();
      }

      public void UpdateText()
      {
        Text = (this as IMailboxContainer).GetStatusText();
      }

      object IMailboxContainer.GetMailbox()
      {
        return null;
      }

      string IMailboxContainer.GetStatusText()
      {
        return string.Format(SIPrefixFormat.CurrentInfo, "全{0}件 ({1:F})", Account.Client.MessageCount, Account.Client.TotalSize);
      }
    }

    private interface IMessageContainer : IAccountContainer {
      object Message { get; }
      string UniqueId { get; }
      bool IsDeleted { get; }
      Mail Mail { get; }
      bool CanCancelLoading { get; }

      Task LoadMailAsync(CancellationToken cancellationToken);
      Task SaveAsync(string path, CancellationToken cancellationToken);
    }

    private interface IMessageListViewItem {
      int CompareTo(int column, ListViewItem.ListViewSubItem other);
      void UpdateView();
    }

    private abstract class MessageItemBase : ListViewItem {
      protected static string ToString(MimeType mediaType)
      {
        if (mediaType == null)
          return string.Empty;
        else
          return mediaType.ToString();
      }

      protected static string ToString(DateTimeOffset? dto, string alternativeValue)
      {
        if (dto == null)
          return alternativeValue ?? "-";
        else
          return dto.Value.LocalDateTime.ToString("yyyy-MM-dd HH:mm:ss");
      }
    }

    private class ImapMessageItem : MessageItemBase, IMessageContainer, IMessageListViewItem {
      public ImapAccount Account {
        get; private set;
      }

      AccountBase IAccountContainer.Account {
        get { return Account; }
      }

      public ImapMessageInfo Message {
        get; private set;
      }

      public Mail Mail {
        get; private set;
      }

      public bool IsDeleted {
        get { return Message.IsDeleted; }
      }

      object IMessageContainer.Message {
        get { return Message; }
      }

      public string UniqueId {
        get { return Message.Uid.ToString(); }
      }

      public bool CanCancelLoading {
        get { return true; }
      }

      private readonly IProgress<ImapMessageFetchProgressInfo> progress;

      public ImapMessageItem(ImapAccount account,
                             ImapMessageInfo message,
                             IProgress<ImapMessageFetchProgressInfo> progress)
      {
        this.Account = account;
        this.Message = message;
        this.progress = progress;

        UpdateView();
      }

      public async Task LoadMailAsync(CancellationToken cancellationToken)
      {
        Mail = await Message.ReadAsAsync(Mail.Load, cancellationToken, progress);
      }

      public async Task SaveAsync(string path, CancellationToken cancellationToken)
      {
        await Message.SaveAsync(path, cancellationToken, progress);
      }

      public void UpdateView()
      {
        this.SubItems.Clear();

        this.Text = Message.EnvelopeSubject;
        this.SubItems.Add(Message.Uid.ToString());
        this.SubItems.Add(Message.Length.ToString());
        this.SubItems.Add(ToString(Message.MediaType));
        this.SubItems.Add(ToString(Message.InternalDate, null));
        this.SubItems.Add(ToString(Message.EnvelopeDate, null));
        this.SubItems.Add(ToString(Message.Envelope.From));
        this.SubItems.Add(Message.Flags.ToString());

        if (!Message.IsDeleted && Message.IsRecent)
          ForeColor = Color.Blue;
        else
          ForeColor = Color.Black;

        if (Message.IsDeleted)
          Font = Fonts.StrikeoutFont;
        else if (Message.IsSeen)
          Font = Control.DefaultFont;
        else
          Font = Fonts.BoldFont;
      }

      public int CompareTo(int column, ListViewItem.ListViewSubItem other)
      {
        switch (column) {
          case 1: return (int)(Message.Uid - long.Parse(other.Text));
          case 2: return (int)(Message.Length - long.Parse(other.Text));
          default: return string.Compare(SubItems[column].Text, other.Text);
        }
      }

      private static string ToString(IReadOnlyList<ImapAddress> addresses)
      {
        if (addresses == null || addresses.Count == 0)
          return string.Empty;

        return string.Join(", ", addresses.Select(delegate(ImapAddress address) {
          return string.Format("{0} <{1}@{2}>",
                               address.GetDecodedName(),
                               address.Mailbox,
                               address.Host);
        }));
      }
    }

    private class PopMessageItem : MessageItemBase, IMessageContainer, IMessageListViewItem {
      public PopAccount Account {
        get; private set;
      }

      AccountBase IAccountContainer.Account {
        get { return Account; }
      }

      public PopMessageInfo MessageInfo {
        get; private set;
      }

      public Mail Mail {
        get; private set;
      }

      public bool IsDeleted {
        get { return MessageInfo.IsMarkedAsDeleted; }
      }

      object IMessageContainer.Message {
        get { return MessageInfo; }
      }

      public string UniqueId {
        get { return MessageInfo.UniqueId ?? string.Format("#{0}", MessageInfo.MessageNumber); }
      }

      public bool CanCancelLoading {
        get { return false; }
      }

      private readonly IProgress<PopMessageRetrieveProgressInfo> progress;

      public PopMessageItem(PopAccount account,
                            PopMessageInfo messageInfo,
                            Mail mail,
                            IProgress<PopMessageRetrieveProgressInfo> progress)
      {
        this.Account = account;
        this.MessageInfo = messageInfo;
        this.progress = progress;

        UpdateMessage(mail);
      }

      public void UpdateMessage(Mail mail)
      {
        this.Mail = mail;

        this.Text = mail.Subject ?? string.Empty;
        this.SubItems.Add(MessageInfo.UniqueId ?? "(Unique-IDを取得できません)");
        this.SubItems.Add(MessageInfo.Length.ToString());
        this.SubItems.Add(ToString(mail.MimeType));
        this.SubItems.Add("-");
        this.SubItems.Add(ToString(mail.Date, mail.DateString));
        this.SubItems.Add(mail.From ?? string.Empty);
        this.SubItems.Add("-");
      }

      public void UpdateView()
      {
        if (MessageInfo.IsMarkedAsDeleted)
          Font = Fonts.StrikeoutFont;
        else
          Font = Control.DefaultFont;
      }

      public int CompareTo(int column, ListViewItem.ListViewSubItem other)
      {
        switch (column) {
          case 2: return (int)(MessageInfo.Length - long.Parse(other.Text));
          default: return string.Compare(SubItems[column].Text, other.Text);
        }
      }

      public async Task LoadMailAsync(CancellationToken cancellationToken)
      {
        Mail = await MessageInfo.ReadAsAsync(Mail.Load, cancellationToken, progress);
      }

      public async Task SaveAsync(string path, CancellationToken cancellationToken)
      {
        await MessageInfo.SaveAsync(path, cancellationToken, progress);
      }
    }

    private class SectionDropDownList : FormattingDropDownList<SectionDropDownListItem> {
      public SectionDropDownList()
      {
        StringFormatter = SectionDropDownListItem.Format;
      }
    }

    private class SectionDropDownListItem : IMessageContainer {
      public AccountBase Account {
        get; private set;
      }

      public ImapMessagePartInfo MessagePart {
        get; private set;
      }

      public Mail Mail {
        get; private set;
      }

      object IMessageContainer.Message {
        get { return MessagePart.Message; }
      }

      public bool IsDeleted {
        get { return MessagePart.Message.IsDeleted; }
      }

      public string UniqueId {
        get { return MessagePart.Message.Uid.ToString(); }
      }

      public bool CanCancelLoading {
        get { return true; }
      }

      private readonly IProgress<ImapMessageFetchProgressInfo> progress;

      public SectionDropDownListItem(AccountBase account,
                                     ImapMessagePartInfo messagePart,
                                     IProgress<ImapMessageFetchProgressInfo> progress)
      {
        this.Account = account;
        this.MessagePart = messagePart;
        this.progress = progress;
      }

      public async Task LoadMailAsync(CancellationToken cancellationToken)
      {
        Mail = await MessagePart.ReadAsAsync(Mail.Load, cancellationToken, progress);
      }

      public async Task SaveAsync(string path, CancellationToken cancellationToken)
      {
        await MessagePart.SaveAsync(path, cancellationToken, progress);
      }

      public static string Format(SectionDropDownListItem item)
      {
        var specifier = item.MessagePart.Specifier;

        return string.Format("{0} ({1}) <{2}>",
                             specifier.Length == 0 ? "body" : "section " + specifier,
                             item.MessagePart.ContentType,
                             item.MessagePart.Url.LocalPath);
      }
    }

    private class QuotaResourceToolStripItem : ToolStripItem {
      public string QuotaRoot {
        get { return quotaRoot; }
        set
        {
          if (quotaRoot != value) {
            quotaRoot = value;
            OnChanged();
          }
        }
      }

      public ImapQuotaResource Resource {
        get { return resource; }
        set
        {
          if (resource != value) {
            resource = value;
            OnChanged();
          }
        }
      }

      private double UsagePercent {
        get
        {
          if (resource == null || resource.Limit == 0L)
            return 0.0;
          else
            return resource.Usage / (double)resource.Limit;
        }
      }

      public QuotaResourceToolStripItem()
      {
        AutoSize = true;
        AutoToolTip = true;
      }

      private void OnChanged()
      {
        if (resource == null) {
          Text =
          ToolTipText = string.Empty;
        }
        else {
          Text = string.Format("{0}: {1:P1}", resource.Name, UsagePercent);

          if (string.Equals(resource.Name, "STORAGE", StringComparison.OrdinalIgnoreCase))
            ToolTipText = string.Format(SIPrefixFormat.CurrentInfo, "{0:f}/{1:f}", resource.Usage * 1024L, resource.Limit * 1024L);
          else
            ToolTipText = string.Format("{0}/{1}", resource.Usage, resource.Limit);

          ToolTipText = string.Format("[{0}] {1} ({2})", quotaRoot, Text, ToolTipText);
        }

        Invalidate();
      }

      private double GetUsage()
      {
        return (resource.Limit == 0L) ? 0.0 : resource.Usage / (double)resource.Limit;
      }

      protected override void OnPaint(PaintEventArgs e)
      {
        base.OnPaint(e);

        if (resource == null)
          return;

        var g = e.Graphics;
        var bounds = Bounds;

        bounds.X = 0;

        g.FillRectangle(Brushes.DarkGray, bounds);

        using (var b = new LinearGradientBrush(new Point(bounds.Left, bounds.Top),
                                               new Point(bounds.Right, bounds.Top),
                                               Color.Green,
                                               Color.Red)) {
          var blend = new ColorBlend();

          blend.Colors = new Color[] {
            Color.Green,
            Color.Yellow,
            Color.Red,
          };
          blend.Positions = new float[] {
            0.0f,
            0.5f,
            1.0f,
          };

          b.InterpolationColors = blend;

          var fillBounds = bounds;

          fillBounds.Width = (int)(fillBounds.Width * UsagePercent);

          g.FillRectangle(b, fillBounds);
        }

        using (var sf = StringFormat.GenericDefault.Clone() as StringFormat) {
          sf.Alignment = StringAlignment.Center;
          sf.LineAlignment = StringAlignment.Center;

          g.DrawString(Text, Font, Brushes.Black, bounds, sf);
        }

        g.DrawRectangle(Pens.Black, bounds);
      }

      private string quotaRoot;
      private ImapQuotaResource resource;
    }

    private string mimeTypesFile;
    private bool verifyServerCertificates;

    private ImapAccount[] imapAccounts;
    private PopAccount[] popAccounts;
    private AccountBase[] allAccounts;
    private readonly Dictionary<ImapMailboxInfo, ImapMailboxNode> imapMailboxNodeMap = new Dictionary<ImapMailboxInfo, ImapMailboxNode>();
    private readonly Dictionary<PopAccount, PopMailboxNode> popMailboxNodeMap = new Dictionary<PopAccount, PopMailboxNode>();
    private StringBuilder statusStringBuffer = new StringBuilder(10240);
    private ToolStripSeparator contextMenuDummyItem = new ToolStripSeparator();
    private Progress<ImapMessageFetchProgressInfo> imapDownloadProgress;
    private Progress<PopMessageRetrieveProgressInfo> popDownloadProgress;
    private DateTime progressLastUpdate = DateTime.Now;
  }
}
