using System;
using System.Collections.Generic;
using System.Net;
using System.Threading.Tasks;

using Smdn.Net.Pop3;
using Smdn.Net.Pop3.Client;

namespace Smdn.Applications.SimpleMailer {
  public class PopAccount : AccountBase {
    public PopClient Client {
      get { return client; }
    }

    public bool ReadOnly {
      get; private set;
    }

    public override bool IsReady {
      get { return !(client == null || client.IsBusy); }
    }

    public PopAccount(string name, Uri serverUri, ICredentialsByHost credential, int timeout, bool useTlsIfAvailable)
      : base(name, serverUri)
    {
      this.client = new PopClient(serverUri);
      this.credential = credential;

      this.client.Profile.Timeout = timeout;
      this.client.Profile.UseTlsIfAvailable = useTlsIfAvailable;
    }

    public override Task EnsureConnected()
    {
      if (client.IsConnected) {
        var tcs = new TaskCompletionSource<object>();

        tcs.SetResult(null);

        return tcs.Task;
      }

      return client.ConnectAsync(credential).ContinueWith((_) => {;
        this.credential = null;
      }, TaskContinuationOptions.OnlyOnRanToCompletion);
    }

    private PopClient client;
    private ICredentialsByHost credential;
  }
}
