﻿using System;
using System.Net.Security;
using System.Security.Cryptography.X509Certificates;
using System.Text;
using System.Windows.Forms;

namespace Smdn.Applications.SimpleMailer {
  public partial class ValidateCertDialog : Form {
    public enum Result {
      Accept,
      AcceptOnlyOnce,
      Abort,
    }

    public ValidateCertDialog()
    {
      InitializeComponent();
    }

    public Result ValidateCert(IWin32Window owner, X509Certificate certificate, X509Chain chain, SslPolicyErrors sslPolicyErrors)
    {
      labelMessage.Text = string.Format("サーバ証明書に問題があります。　この証明書を受け入れますか?{0}{0}(SslPolicyErrors: {1})", Environment.NewLine, sslPolicyErrors);

      var detail = new StringBuilder();

      detail.AppendFormat(certificate.ToString(false));
      detail.AppendLine();

      if (chain != null) {
        foreach (var element in chain.ChainElements) {
          detail.AppendFormat("発行先: {0}", element.Certificate.Subject).AppendLine();
          detail.AppendFormat("発行者: {0}", element.Certificate.Issuer).AppendLine();
          detail.AppendFormat("有効期間: {0} から {1}",
            element.Certificate.GetEffectiveDateString(),
            element.Certificate.GetExpirationDateString()
          ).AppendLine();
          detail.AppendLine();

          foreach (var status in element.ChainElementStatus) {
            detail.AppendFormat("Status: {0}", status.Status).AppendLine();
            detail.AppendFormat("StatusInformation: {0}", status.StatusInformation).AppendLine();
          }
        }
      }

      textBoxCertDetail.Text = detail.ToString();
      textBoxCertDetail.Select(0, 0);

      switch (ShowDialog(owner)) {
        case DialogResult.OK:
          return Result.Accept;

        case DialogResult.Cancel:
          return Result.Abort;

        case DialogResult.Retry:
        default:
          return Result.AcceptOnlyOnce;
      }
    }
  }
}
